"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.WhereBuilder = void 0;
const typeorm_1 = require("typeorm");
const sql_comparison_builder_1 = require("./sql-comparison.builder");
/**
 * @internal
 * Builds a WHERE clause from a Filter.
 */
class WhereBuilder {
    constructor(sqlComparisonBuilder = new sql_comparison_builder_1.SQLComparisonBuilder()) {
        this.sqlComparisonBuilder = sqlComparisonBuilder;
    }
    /**
     * Builds a WHERE clause from a Filter.
     * @param where - the `typeorm` WhereExpressionBuilder
     * @param filter - the filter to build the WHERE clause from.
     * @param relationNames - the relations tree.
     * @param alias - optional alias to use to qualify an identifier
     */
    build(where, filter, relationNames, alias) {
        const { and, or, exists, notExists } = filter;
        if (and && and.length) {
            this.filterAnd(where, and, relationNames, alias);
        }
        if (or && or.length) {
            this.filterOr(where, or, relationNames, alias);
        }
        if (exists) {
            this.filterExistsNotExists(where, exists, relationNames, alias, true);
        }
        if (notExists) {
            this.filterExistsNotExists(where, notExists, relationNames, alias, false);
        }
        return this.filterFields(where, filter, relationNames, alias);
    }
    /**
     * ANDs multiple filters together. This will properly group every clause to ensure proper precedence.
     *
     * @param where - the `typeorm` WhereExpressionBuilder
     * @param filters - the array of filters to AND together
     * @param relationNames - the relations tree.
     * @param alias - optional alias to use to qualify an identifier
     */
    filterAnd(where, filters, relationNames, alias) {
        return where.andWhere(new typeorm_1.Brackets((qb) => filters.reduce((w, f) => qb.andWhere(this.createBrackets(f, relationNames, alias)), qb)));
    }
    /**
     * ORs multiple filters together. This will properly group every clause to ensure proper precedence.
     *
     * @param where - the `typeorm` WhereExpressionBuilder
     * @param filter - the array of filters to OR together
     * @param relationNames - the relations tree.
     * @param alias - optional alias to use to qualify an identifier
     */
    filterOr(where, filter, relationNames, alias) {
        return where.andWhere(new typeorm_1.Brackets((qb) => filter.reduce((w, f) => qb.orWhere(this.createBrackets(f, relationNames, alias)), qb)));
    }
    /**
     * Wraps a filter in brackes to ensure precedence.
     * ```
     * {a: { eq: 1 } } // "(a = 1)"
     * {a: { eq: 1 }, b: { gt: 2 } } // "((a = 1) AND (b > 2))"
     * ```
     * @param filter - the filter to wrap in brackets.
     * @param relationNames - the relations tree.
     * @param alias - optional alias to use to qualify an identifier
     */
    createBrackets(filter, relationNames, alias) {
        return new typeorm_1.Brackets((qb) => this.build(qb, filter, relationNames, alias));
    }
    /**
     * Creates field comparisons from a filter. This method will ignore and/or properties.
     * @param where - the `typeorm` WhereExpressionBuilder
     * @param filter - the filter with fields to create comparisons for.
     * @param relationNames - the relations tree.
     * @param alias - optional alias to use to qualify an identifier
     */
    filterExistsNotExists(where, filter, relationMap, alias, isExists = true) {
        var _a, _b;
        if (!relationMap.metadata || !alias) {
            return where;
        }
        // `where` is in fact SelectQueryBuilder but it is not typed as such hence the cast
        // eslint-disable-next-line @typescript-eslint/no-explicit-any
        const qb = where;
        let sqb = qb.subQuery().select('1');
        if (relationMap.metadata.relationType === 'one-to-one' || relationMap.metadata.relationType === 'one-to-many') {
            const [joinColumn] = (_b = (_a = relationMap.metadata.inverseRelation) === null || _a === void 0 ? void 0 : _a.joinColumns) !== null && _b !== void 0 ? _b : [];
            if (!joinColumn) {
                return where;
            }
            sqb
                .from(relationMap.metadata.inverseEntityMetadata.tableName, alias)
                .andWhere(`${alias}.${joinColumn.databaseName} = ${qb.alias}.id`);
        }
        else if (relationMap.metadata.relationType === 'many-to-many' && relationMap.metadata.junctionEntityMetadata) {
            const joinTableName = relationMap.metadata.joinTableName;
            const [ownerColumn] = relationMap.metadata.junctionEntityMetadata.ownerColumns;
            const [inverseColumn] = relationMap.metadata.junctionEntityMetadata.inverseColumns;
            if (!ownerColumn || !inverseColumn) {
                return where;
            }
            const joinAlias = `${qb.alias}_${alias}`;
            sqb
                .from(joinTableName, joinAlias)
                .innerJoin(relationMap.metadata.inverseEntityMetadata.tableName, alias, `${joinAlias}.${inverseColumn.databaseName} = ${alias}.id`)
                .andWhere(`${joinAlias}.${ownerColumn.databaseName} = ${qb.alias}.id`);
        }
        else {
            return where;
        }
        const subWhere = new WhereBuilder();
        sqb = subWhere.build(sqb, filter, relationMap, alias);
        return where.andWhere(`${isExists ? 'EXISTS' : 'NOT EXISTS'} (${sqb.getQuery()})`, qb.getParameters());
    }
    /**
     * Creates field comparisons from a filter. This method will ignore and/or properties.
     * @param where - the `typeorm` WhereExpressionBuilder
     * @param filter - the filter with fields to create comparisons for.
     * @param relationNames - the relations tree.
     * @param alias - optional alias to use to qualify an identifier
     */
    filterFields(where, filter, relationNames, alias) {
        return Object.keys(filter).reduce((w, field) => {
            if (field !== 'and' && field !== 'or' && field !== 'exists' && field !== 'notExists') {
                return this.withFilterComparison(where, field, this.getField(filter, field), relationNames, alias);
            }
            return w;
        }, where);
    }
    getField(obj, field) {
        return obj[field];
    }
    withFilterComparison(where, field, cmp, relationNames, alias) {
        if (relationNames.nested[field]) {
            return this.withRelationFilter(where, field, cmp, relationNames.nested[field]);
        }
        return where.andWhere(new typeorm_1.Brackets((qb) => {
            const opts = Object.keys(cmp);
            const sqlComparisons = opts.map((cmpType) => this.sqlComparisonBuilder.build(field, cmpType, cmp[cmpType], alias));
            sqlComparisons.map(({ sql, params }) => qb.orWhere(sql, params));
        }));
    }
    withRelationFilter(where, field, cmp, relationNames) {
        return where.andWhere(new typeorm_1.Brackets((qb) => {
            const relationWhere = new WhereBuilder();
            return relationWhere.build(qb, cmp, relationNames, field);
        }));
    }
}
exports.WhereBuilder = WhereBuilder;
//# sourceMappingURL=where.builder.js.map